#define BUFFER_SIZE 13
#define PACKET_UNKNOWN 1
#define PACKET_BALLDATA 2
#define PACKET_BUTTONDATA 3

unsigned short sensitivity_chart( unsigned short value );

inline unsigned short mshl( unsigned short b, unsigned short mask, unsigned short shift )
{
  return (b & mask) << shift;
}

inline unsigned short mshr( unsigned short b, unsigned short mask, unsigned short shift )
{
  return (b & mask) >> shift;
}

inline short reshift( unsigned short b )
{
  return ((((short) b) << 6) >> 6);
}

// this has to be straight C++, darn you arduino
class OrbBuffer
{
public:
  unsigned char buffer[BUFFER_SIZE];
  unsigned short cursor;
  short physical_axes[6];
  unsigned char physical_buttons;
  unsigned short logical_axes[6];
  unsigned short logical_buttons;
  unsigned short axis_map[6];
  bool chording;
  // polarity is by LOGICAL axis
  unsigned short polarity[6];
  bool has_changed;
  OrbBuffer()
  {
    reset();
    physical_axes[0]=physical_axes[1]=physical_axes[2]=physical_axes[3]=physical_axes[4]=physical_axes[5]=0;
    physical_buttons = 0;
    logical_axes[0]=logical_axes[1]=logical_axes[2]=logical_axes[3]=logical_axes[4]=logical_axes[5]=0;
    logical_buttons = 0;
    set_axis_map(0,1,2,3,4,5);
    set_polarity( 1,1,1,1,1,1 );
    chording = true;
  }

  // axis-map is which physical axis belongs to this logical axis, ie axis_map[0]=1 means 
  // logical axis 0 is taken from physical axis 1
  void set_axis_map( unsigned short a0, unsigned short a1, unsigned short a2,
  unsigned short a3, unsigned short a4, unsigned short a5 )
  {
    axis_map[0] = a0;
    axis_map[1] = a1;
    axis_map[2] = a2;
    axis_map[3] = a3;
    axis_map[4] = a4;
    axis_map[5] = a5;
  }

  void set_polarity( int a0, int a1, int a2, int a3, int a4, int a5 )
  {
    polarity[0] = a0;
    polarity[1] = a1;
    polarity[2] = a2;
    polarity[3] = a3;
    polarity[4] = a4;
    polarity[5] = a5;
  }

  void reset( void )
  {
    cursor = 0;
  }

  // call this once you have taken note of the changed internal state of the
  // orb and sent the data out
  void log_change( void )
  {
    has_changed = false;
  }

  unsigned short current_packet_type( void )
  {
    if ( cursor == 0 )
    {
      return PACKET_UNKNOWN;
    }
    else
    {
      switch (buffer[0])
      {
      case 'K' :
        return PACKET_BUTTONDATA;
        break;
      case 'D' :
        return PACKET_BALLDATA;
        break;
      }
    }
  }

  bool in_packet( void )
  {
    return (current_packet_type() != PACKET_UNKNOWN );
  }

  void process_buttondata( void )
  {
    // buffer[2] now has data in the form 1<rezero><F><E><D><C><B><A>
    physical_buttons = buffer[2] & 63;  
  }

  void process_balldata( void )
  {
    // D<button status><D1><D2><D3><D4><D5><D6><D7><D8><D9><xor>
    // from elsewhere:
    // <D1> to <D9> are data bytes containing the force and torque components. 
    // These values are generated by first packing the six 10 bit values then 
    // xoring them with the characters "SpaceWare". Packed data has the form: 
    // 1XXX XXXX 1XXX YYYY 1YYY YYYZ 1ZZZ ZZZZ1ZZx 
    // xxxx 1xxx xxyy 1yyy yyyy 1yzz zzzz1zzz z000 
    // where XXXXXXXXXX , YYYYYYYYYY , ZZZZZZZZZZ are the components of force. 
    // xxxxxxxxxx , yyyyyyyyyy , zzzzzzzzzz are the components of torque.
    // ...
    // Note the "at rest" packet, the packet with all force and torque components at zero, 
    // will contain the word "SpaceWare".

    // great; so first let's xor with "D.SpaceWare"
    unsigned short i;
    char* string_to_xor = "SpaceWare";
    for ( i = 0; i < 9; ++i )
    {
      buffer[i+2] ^= string_to_xor[i];
    }
    // now process button char
    physical_buttons = buffer[1] & 63;
    // now process ball components
    physical_axes[0] = mshl( buffer[2], 0x7F, 3) | mshr( buffer[3], 0x70, 4 ); //(((buffer[2] & 0x7F) << 3) | ((buffer[3] & 0x70) >> 4));
    physical_axes[1] = mshl( buffer[3], 0x0F, 6) | mshr( buffer[4], 0x7E, 1 ); //(((buffer[3] & 0x0F) << 6) | ((buffer[4] & 0x7E) >> 1));
    physical_axes[2] = mshl( buffer[4], 0x01, 9) | mshl( buffer[5], 0x7F, 2 ) | mshr( buffer[6], 0x60, 5 );    //(((buffer[4] & 0x01) << 9) | ((buffer[5] & 0x7F) << 2) | ((buffer[6] & 0x60) >> 5));
    physical_axes[3] = mshl( buffer[6], 0x1F, 5) | mshr( buffer[7], 0x7C, 2 ); //(((buffer[6] & 0x1F) << 5) | ((buffer[7] & 0x7C) >> 2));
    physical_axes[4] = mshl( buffer[7], 0x03, 8) | mshl( buffer[8], 0x7F, 1 ) | mshr(buffer[9], 0x40, 6);    //(((buffer[7] & 0x03) << 8) | ((buffer[8] & 0x7F) << 1) | ((buffer[9] & 0x40) >> 6));
    physical_axes[5] = mshl( buffer[9], 0x3f, 4) | mshr( buffer[10], 0x78, 3 ); //(((buffer[9]  & 0x3F) << 4) | ((buffer[10] & 0x78) >> 3));

    //ok, now we have a bunch of signed 10-bit integers.  That means bit 10 is our sign bit
    //and we need to move it to bit 16
    for ( int i = 0; i < 6; ++i )
    {
      physical_axes[i] = reshift( physical_axes[i] );
    }

  }



  unsigned short scaled_axis_value( short physical_val, int polarity )
  {
    switch( polarity )
    {
    case 1 : 
      return sensitivity_chart( physical_val + 512 );
      break;
    case 0 :
      return 0;
      break;
    case -1 :
      return sensitivity_chart( 512 - physical_val );
      break;
    }
  }

  void map_physical_to_logical()
  {
    // work chording here
    if ( chording )
    {
      int chord_page = physical_buttons & 3;
      logical_buttons = ( physical_buttons >> 2 ) << ( 4*chord_page);
    }
    else
    {
      logical_buttons = physical_buttons & 63;
    }
    // work mapping here
    for ( unsigned short i = 0; i < 6; ++i )
    {
      logical_axes[i] = physical_axes[axis_map[i]] + 512; //scaled_axis_value( physical_axes[ axis_map[i] ], polarity[i] );
    }
  }

  bool finalize_packet( void )
  {
    switch (current_packet_type())
    {
    case PACKET_BUTTONDATA :
      process_buttondata();
      break;
    case PACKET_BALLDATA :
      process_balldata();
      break;

    }
    map_physical_to_logical();
    has_changed = true;
  }

  bool start_new_packet( unsigned char b )
  {
    buffer[0] = b;
    cursor = 1;
  }

  bool add_byte( unsigned char b )
  {
    switch ( b )
    {
    case 'D' :
    case 'K' :
      start_new_packet(b);
      break;
    case 13 :
      finalize_packet();
    default :
      if (in_packet() && (cursor < BUFFER_SIZE))
      {
        buffer[cursor] = b;
        cursor++;
      }
    }

  }


};
